# PRG - Exam Test #

This is the assignment for the PRG exam test. Follow the instructions precisely and do not add
any additional code to the modules implemented (not even print statements) as they will be imported
and evaluated automatically. Syntactic errors in a module result in 0 points from the assignment,
as all automatic tests will fail on that module.

The modules (packages) you implement should be packed in a zip file (files should be in the root of
the zip file, zipping directly the files, not the containing folder) and uploaded to BRUTE. For
every module, class and function you implement, do not forget to add a docstring. Code quality
(mainly following naming conventions and writing docstrings) is part of the grading of your
solution.


## Task 1 [2p] ##

Create a module named "type_conversion". In this module, implement a function "int" that works on
strings and converts a string containing a number with a decimal part to a string containing
a number without the decimal part ("3.3" -> "3"). In the docstring of the function, besides the
functionality description, explain in which scope the function is defined, if the same identifier is
in any other scope, and which one will take precedence in this module.


## Task 2 [5p] ##

Create a module named "task2_functions" and implement three functions according to the following
unit tests. You can put the unit tests (from below) to the same module.

```
import unittest

class TestFunctions(unittest.TestCase):

    def test_reformat_date(self):
        # [2p]
        # DD, MM, YYYY stands for day, month and year respectively
        self.assertEqual("31.12.2020", reformat_date("20201231", "DD.MM.YYYY"))
        self.assertEqual("01/12/2020", reformat_date("20200112", "MM/DD/YYYY"))
        self.assertEqual("2021-01-01", reformat_date("20210101", "YYYY-MM-DD"))
        self.assertEqual("2019", reformat_date("20190521", "YYYY"))

    def test_exclude_worst(self):
        # [1.5p]
        self.assertEqual([4, 6, 10], exclude_worst([10, 4, 1, 6]))
        self.assertEqual([6, 10], exclude_worst([10, 4, 1, 6], exclude=2))
        self.assertEqual([10], exclude_worst([10, 4, 1, 6], exclude=3))

    def test_is_pin_valid(self):
        # [1.5p]
        self.assertTrue(is_pin_valid("4412"))
        self.assertFalse(is_pin_valid("12442"))
        self.assertFalse(is_pin_valid("122"))
        self.assertFalse(is_pin_valid("12ac"))
        self.assertFalse(is_pin_valid("12.?"))
```


## Task 3 [6p] ##

Create a module named "list_io", which will contain two functions. These two function will save and
load a list of strings in a text file. The format of the file is up to you, only make sure that
the load function can load what the save function has saved.

1. [2p] Implement function save_list, which takes two arguments - a list of strings and a file
    path, and will store the list under given path.
2. [2p] In function save_list, if the folder where the data file should be created, does not exist,
    create it. Create only the containing folder - for path "/foo/bar/file.txt", only "bar" is
    created if it does not exist.
3. [2p] Implement function load_list, which takes a single argument, a file path, and returns
    a list loaded. If the file does not exist, return an empty list.


## Task 4 [6p] ##

Create a package containing a single module. In this module, implement two functions operating on
strings.

1. [1p] Choose the name of the package, module and functions appropriately based on the
    functionality defined below.
2. [1p] The first function is named 'reverse_words' and takes a single argument, space-separated
    words, and returns the words in reverse (e.g. "white fast horse" -> "horse fast white").
3. [2p] The second function takes a single argument, a string containing words and spaces, and
    ensures that words are separated only with a single space, replacing consecutive spaces with
    a single one and removing the leading and trailing spaces (e.g. "  white      horse " ->
    "white horse").
4. [2p] On the same level as the package (top-level folder), create a module containing unit tests
    for the implemented module.


## Task 5 [6p] ##

Create a module named "expression".

1. [1p] Implement class MathematicalExpression that stores a mathematical expression in a text format.
1. [2p] Implement support for the plus operator, where the first argument is always
    MathematicalExpression instance and second argument can be either numeric (int, float) or also
    MathematicalExpression instance. The result of the addition is a new MathematicalExpression
    instance.
2. [1p] Implement the equality operator - two MathematicalExpression instances are equal if and only
    if the expressions are equal (not only the results). Spaces in the expressions do not matter.
3. [2p] Implement a class IntegerExpression that inherits from MathematicalExpression and which
    expression can only contain integer numbers (not floating point numbers, i.e. "3+3" not
    "3.3+3"). Ensure this by raising an appropriate exception. Try to reuse as much code from
    MathematicalExpression as possible.

MathematicalExpression example:

```
>>> MathematicalExpression("2 + 3") == MathematicalExpression("2+3")
True
>>> MathematicalExpression("2 + 3 + 4 + 1 + 0.5") == MathematicalExpression("2+3") + MathematicalExpression("4") + 1 + 0.5
True
```


## Epilogue ##

At the end of the test, the following files (folders) should be present in your zip file. They
should be in the root of the zip file, not in a nested folder.

- [folder of chosen name] - task 4
- expression.py - task 5
- list_io.py - task 3
- task2_functions.py - task 2
- test_[module of chosen name].py - task 4
- type_conversions.py - task 1

